@@ -6,12 +6,13 @@ use std::{
     fs::{self, File},
     io::Write,
     path::{Path, PathBuf},
+    collections::BTreeMap,
 };
 use irc::client::data::Config as IrcConfig;
 use failure::{Error, bail};
 use directories::{BaseDirs, ProjectDirs};
 use serde_derive::{Serialize, Deserialize};
-use log::{info, warn};
+use log::info;
 
 use crate::{
     VERSION,
@@ -88,6 +89,7 @@ pub struct Parameters {
     pub status_channels: Vec<String>,
     pub nick_response_str: String,
     pub reconnect_timeout: u64,
+    pub ignore_nicks: Vec<String>,
 }
 
 impl Default for Parameters {
@@ -97,6 +99,7 @@ impl Default for Parameters {
             status_channels: vec![],
             nick_response_str: "".to_string(),
             reconnect_timeout: 10,
+            ignore_nicks: vec![],
         }
     }
 }
@@ -155,13 +158,17 @@ pub struct Conf {
     pub database: Database,
     #[serde(rename = "connection")]
     pub client: IrcConfig,
+    #[serde(skip)]
+    pub path: Option<PathBuf>,
 }
 
 impl Conf {
     /// load configuration TOML from a file
     pub fn load(path: impl AsRef<Path>) -> Result<Self, Error> {
         let conf = fs::read_to_string(path.as_ref())?;
-        let conf: Conf = toml::de::from_str(&conf)?;
+        let mut conf: Conf = toml::de::from_str(&conf)?;
+        // insert the path the config was loaded from
+        conf.path = Some(path.as_ref().to_path_buf());
         Ok(conf)
     }
 
@@ -214,10 +221,39 @@ impl Default for Conf {
                 user_info: Some("Feed me URLs.".to_string()),
                 ..IrcConfig::default()
             },
+            path: None,
         }
     }
 }
 
+#[derive(Serialize, Deserialize, Clone)]
+pub struct ConfSet {
+    #[serde(flatten)]
+    pub configs: BTreeMap<String, Conf>,
+}
+
+impl ConfSet {
+    /// load configuration TOML from a file
+    pub fn load(path: impl AsRef<Path>) -> Result<Self, Error> {
+        let conf_string = fs::read_to_string(path.as_ref())?;
+        let mut conf_set: ConfSet = toml::de::from_str(&conf_string)?;
+
+        // populate path field of all configs
+        conf_set.configs
+            .iter_mut()
+            .for_each(|(_, c)| c.path = Some(path.as_ref().to_path_buf()));
+
+        Ok(conf_set)
+    }
+
+    /// write configuration to a file
+    pub fn write(&self, path: impl AsRef<Path>) -> Result<(), Error> {
+        let mut file = File::create(path)?;
+        file.write_all(toml::ser::to_string(&self)?.as_bytes())?;
+        Ok(())
+    }
+}
+
 /// Run-time configuration data.
 #[derive(Default, Clone)]
 pub struct Rtd {
@@ -231,7 +267,6 @@ pub struct Rtd {
 
 #[derive(Default, Clone)]
 pub struct Paths {
-    pub conf: PathBuf,
     pub db: Option<PathBuf>,
 }
 
@@ -240,9 +275,9 @@ impl Rtd {
         Rtd::default()
     }
 
-    /// Set the configuration file path.
-    pub fn conf(mut self, path: &Path) -> Self {
-        self.paths.conf = expand_tilde(path);
+    /// Set the configuration
+    pub fn conf(mut self, c: Conf) -> Self {
+        self.conf = c;
         self
     }
 
@@ -280,20 +315,7 @@ impl Rtd {
 
     /// Load the configuration file and return an Rtd.
     pub fn load(mut self) -> Result<Self, Error> {
-        ensure_parent_dir(&self.paths.conf)?;
-
-        // create a default-valued config if it doesn't exist
-        if !self.paths.conf.exists() {
-            info!("Configuration `{}` doesn't exist, creating default",
-                self.paths.conf.to_str().unwrap());
-            warn!("You should modify this file to include a useful IRC \
-                configuration");
-            Conf::default().write(&self.paths.conf)?;
-        }
-
-        // load config file
-        self.conf = Conf::load(&self.paths.conf)?;
-
+        // get a database path
         self.paths.db = self.get_db_path().map(|p| expand_tilde(&p));
 
         if let Some(dp) = &self.paths.db {
@@ -338,7 +360,7 @@ impl Rtd {
     }
 }
 
-fn ensure_parent_dir(file: &Path) -> Result<bool, Error> {
+pub fn ensure_parent_dir(file: &Path) -> Result<bool, Error> {
     let without_path = file.components().count() == 1;
 
     match file.parent() {
@@ -368,10 +390,26 @@ pub fn find_configs_in_dir(dir: &Path) -> Result<impl Iterator<Item = PathBuf>,
     Ok(fs::read_dir(dir)?
         .flatten()
         .map(|e| e.path())
-        .filter(|e| !e.is_dir() && Conf::load(e).is_ok())
+        .filter(|e| !e.is_dir() && (Conf::load(e).is_ok() || ConfSet::load(e).is_ok()))
         .take(32))
 }
 
+/// Take a vector of paths to either configurations, or configuration sets,
+/// and return a vector of configurations
+pub fn load_flattened_configs(paths: Vec<PathBuf>) -> Vec<Conf> {
+    let mut configs: Vec<Conf> = paths.iter()
+        .filter_map(|p| Conf::load(p).ok())
+        .collect();
+
+    let mut set_configs: Vec<Conf> = paths.into_iter()
+        .filter_map(|p| ConfSet::load(p).ok())
+        .flat_map(|s| s.configs.values().cloned().collect::<Vec<Conf>>())
+        .collect();
+
+    configs.append(&mut set_configs);
+    configs
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
@@ -382,11 +420,9 @@ mod tests {
 
     #[test]
     /// test that the example configuration file parses without error
-    fn load_example_conf() {
-        Rtd::new()
-            .conf(&PathBuf::from("example.config.toml"))
-            .load()
-            .unwrap();
+    fn load_example_configs() {
+        Conf::load(&PathBuf::from("example.config.toml")).unwrap();
+        ConfSet::load(&PathBuf::from("example.multi.config.toml")).unwrap();
     }
 
     #[test]
@@ -394,10 +430,8 @@ mod tests {
         let tmp_dir = tempdir().unwrap();
         let cfg_path = tmp_dir.path().join("config.toml");
 
-        Rtd::new()
-            .conf(&cfg_path)
-            .load()
-            .unwrap();
+        let conf = Conf::default();
+        conf.write(&cfg_path).unwrap();
 
         let example = fs::read_to_string("example.config.toml").unwrap();
         let written = fs::read_to_string(cfg_path).unwrap();
@@ -407,6 +441,36 @@ mod tests {
             .for_each(|(a, b)| assert_eq!(a, b));
     }
 
+    fn get_test_confset() -> ConfSet {
+        let mut confset = ConfSet {
+            configs: BTreeMap::new()
+        };
+
+        let mut conf = Conf::default();
+        conf.network.name = String::from("foo");
+        confset.configs.insert("foo".to_string(), conf.clone());
+        conf.network.name = String::from("bar");
+        confset.configs.insert("bar".to_string(), conf);
+
+        confset
+    }
+
+    #[test]
+    fn load_write_default_set() {
+        let tmp_dir = tempdir().unwrap();
+        let cfg_path = tmp_dir.path().join("config.toml");
+
+        let conf = get_test_confset();
+        conf.write(&cfg_path).unwrap();
+
+        let example = fs::read_to_string("example.multi.config.toml").unwrap();
+        let written = fs::read_to_string(cfg_path).unwrap();
+
+        example.lines()
+            .zip(written.lines())
+            .for_each(|(a, b)| assert_eq!(a, b));
+    }
+
     #[test]
     fn sqlite_path_explicit() {
         let tmp_dir = tempdir().unwrap();
@@ -418,8 +482,9 @@ mod tests {
         cfg.database.db_type = DbType::Sqlite;
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .db(Some(&db_path))
             .load()
             .unwrap();
@@ -440,8 +505,9 @@ mod tests {
         cfg.database.path = Some(db_path_cfg.to_str().unwrap().to_string());
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .db(Some(&db_path))
             .load()
             .unwrap();
@@ -461,8 +527,9 @@ mod tests {
         cfg.database.path = Some("".to_string());
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .db(Some(&db_path))
             .load()
             .unwrap();
@@ -480,8 +547,9 @@ mod tests {
         cfg.database.db_type = DbType::Sqlite;
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .load()
             .unwrap();
 
@@ -493,8 +561,9 @@ mod tests {
         cfg.network.name = "test_net".to_string();
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .load()
             .unwrap();
 
@@ -505,8 +574,9 @@ mod tests {
         cfg.database.path = Some("".to_string());
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .load()
             .unwrap();
 
@@ -525,8 +595,9 @@ mod tests {
         cfg.database.path = Some(db_path_cfg.to_str().unwrap().to_string());
         cfg.write(&cfg_path).unwrap();
 
+        let conf = Conf::load(&cfg_path).unwrap();
         let rtd = Rtd::new()
-            .conf(&cfg_path)
+            .conf(conf)
             .load()
             .unwrap();
 
@@ -575,12 +646,7 @@ mod tests {
         assert!(result.is_ok());
     }
 
-    #[test]
-    /// test that the example configuration matches default values
-    fn example_conf_data_matches_generated_default_values() {
-        let example = fs::read_to_string("example.config.toml").unwrap();
-        let default = toml::ser::to_string(&Conf::default()).unwrap();
-
+    fn print_diff(example: &str, default: &str) {
         // print diff (on failure)
         println!("Configuration diff (- example, + default):");
         for diff in diff::lines(&example, &default) {
@@ -590,6 +656,31 @@ mod tests {
                 diff::Result::Right(r) => println!("+{}", r)
             }
         }
+    }
+
+    #[test]
+    /// test that the example configuration matches default values
+    fn example_conf_data_matches_generated_default_values() {
+        let example = fs::read_to_string("example.config.toml").unwrap();
+        let default = toml::ser::to_string(&Conf::default()).unwrap();
+
+        print_diff(&example, &default);
+
+        default.lines()
+            .zip(example.lines())
+            .for_each(|(a, b)| assert_eq!(a, b));
+    }
+
+    #[test]
+    /// test that the example configuration matches default values
+    fn example_conf_data_matches_generated_expected_values() {
+        // construct the example
+        let confset = get_test_confset();
+
+        let example = fs::read_to_string("example.multi.config.toml").unwrap();
+        let default = toml::ser::to_string(&confset).unwrap();
+
+        print_diff(&example, &default);
 
         default.lines()
             .zip(example.lines())
@@ -679,11 +770,7 @@ mod tests {
         let mut f = File::create(&cfg_path).unwrap();
         f.write_all(b"[this]\nis = \"valid toml\"").unwrap();
 
-        let rtd = Rtd::new()
-            .conf(&cfg_path)
-            .load();
-
-        assert!(rtd.is_err());
+        assert!(Conf::load(&cfg_path).is_err());
     }
 
     #[test]
@@ -693,10 +780,7 @@ mod tests {
         let mut f = File::create(&cfg_path).unwrap();
         f.write_all(b"[connection]\n[features]\n[parameters]\n").unwrap();
 
-        Rtd::new()
-            .conf(&cfg_path)
-            .load()
-            .unwrap();
+        Conf::load(&cfg_path).unwrap();
     }
 
     #[test]
@@ -715,4 +799,28 @@ mod tests {
         rtd.conf.features.reconnect = true;
         assert!(feat!(rtd, reconnect));
     }
+
+    #[test]
+    fn test_load_flattened_configs() {
+        let tmp_dir = tempdir().unwrap();
+        let mut paths: Vec<PathBuf> = vec![];
+
+        // make 10 normal configuration files
+        for c in 0..10 {
+            let path = tmp_dir.path().join(format!("conf_{}.toml", c));
+            Conf::default().write(&path).unwrap();
+            paths.push(path);
+        }
+
+        // make 10 configuration sets
+        let set = get_test_confset();
+        for c in 0..10 {
+            let path = tmp_dir.path().join(format!("conf_multi_{}.toml", c));
+            set.write(&path).unwrap();
+            paths.push(path);
+        }
+
+        let res = load_flattened_configs(paths);
+        assert_eq!(res.iter().count(), 30);
+    }
 }
